#include <errno.h>
#include <dirent.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <archive.h>
#include <archive_entry.h>

#include "arg.h"

#define PACK_ROOT  (getenv("ROOT")?getenv("ROOT"):"")
#define PACK_DATA  "var/pm"
#define PACK_BUFF_SIZE 8192
#define PACK_SEPARATOR '#'
#define PACK_EXTENSION ".tar.bz2"

struct pack {
	char *path;
	char *name;
	char *version;
};

/* possible actions */
enum {
	ACTION_INVALID     = -1,
	ACTION_INSTALL     = 0,
	ACTION_DELETE      = 1,
	ACTION_UPDATE      = 2,
	ACTION_INSPECT     = 3,
	ACTION_DEFAULT     = ACTION_INSPECT,
};

/* error codes */
enum {
	ERR_INVALID_ACTION = 1,
	ERR_METADATA       = 2,
	ERR_PACK_LOAD      = 3,
	ERR_DELETE         = 4,
	ERR_INSPECT        = 5,
	ERR_UNPACK         = 6,
};

void usage(char *name);
int is_empty(char *dir);
int mkdir_parents(char *dir, mode_t mode);
char *base_name(char *path);

struct pack *pack_load(char *path);
void pack_free(struct pack *p);

int inspect_collision(char *rootfs, struct pack *p);
int inspect_system(int fd, char *datadir);
int inspect_files(int fd, char *datadir, char *packname);
int inspect(char *datadir, char *packname);

int write_metadata(char *datadir, struct pack *pack);
int write_entry(struct archive *a, struct archive *w);
int unpack(char *rootfs, char *datadir, struct pack *p);
int install(char *rootfs, char *datadir, struct pack *p);
int delete_node(char *path);
int delete_content(char *map, size_t size);
int delete(const char *rootfs, const char *datadir, const char *name);

char *argv0;

int verbose = 0;
int overwrite = 0;

void
usage(char *name)
{
	fprintf(stderr, "usage: %s -adfi <pack>\n" , name);
}


/*
 * Returns 0 if a directory is empty, -1 otherwise
 */
int
is_empty(char *path)
{
	DIR *d;
	struct dirent *p;

	if (!(d = opendir(path))) {
		perror(path);
		return -1;
	}

	while ((p = readdir(d))) {
		if (strcmp(p->d_name, ".") && strcmp(p->d_name, "..")) {
			closedir(d);
			return -1;
		}
	}

	closedir(d);
	return 0;
}


/*
 * Recursive mkdir, taken from the ii project
 * http://nion.modprobe.de/blog/archives/357-Recursive-directory-creation.html
 */
int
mkdir_parents(char *path, mode_t mode)
{
        char tmp[PATH_MAX] = "";
        char *p = NULL;
        size_t len;

	snprintf(tmp, sizeof(tmp), "%s", path);
	len = strlen(tmp);
	if(tmp[len - 1] == '/')
		tmp[len - 1] = 0;
	for(p = tmp + 1; *p; p++)
		if(*p == '/') {
			*p = 0;
			mkdir(tmp, mode);
			*p = '/';
		}
	return mkdir(tmp, mode);
}


/*
 * Return a pointer to the basename, or NULL if path ends with '/'
 */
char *
base_name(char *path)
{
	char *b = strrchr(path, '/');
	return b ? b + 1 : path;
}


/*
 * Check for collisions between the filesystem and the tarball
 */
int
inspect_collision(char *rootfs, struct pack *p)
{
	int r = 0;
	char cwd[PATH_MAX] = "";
	struct stat st;
	struct archive *a;
	struct archive_entry *e;

	a = archive_read_new();
	archive_read_support_filter_bzip2(a);
	archive_read_support_format_tar(a);

	r = archive_read_open_filename(a, p->path, 0);
	if (r != ARCHIVE_OK)
		return -1;

	getcwd(cwd, PATH_MAX);
	chdir(rootfs);
	while (archive_read_next_header(a, &e) == ARCHIVE_OK) {
		if (stat(archive_entry_pathname(e), &st) == 0 && !S_ISDIR(st.st_mode)) {
			fprintf(stderr, "%s: file exists\n", archive_entry_pathname(e));
			r = -1;
			break;
		}
		archive_read_data_skip(a);
	}

	archive_read_free(a);
	chdir(cwd);

	return r;
}


/*
 * Write files installed by a pack to a file descriptor
 */
int
inspect_files(int fd, char *datadir, char *packname)
{
	int meta;
	char tmp[PATH_MAX] = "";
	size_t len;

	snprintf(tmp, PATH_MAX, "%s/%s/files", datadir, packname);
	if ((meta = open(tmp, O_RDONLY)) < 0) {
		perror(tmp);
		return -1;
	}

	while ((len = read(meta, tmp, PATH_MAX)) > 0) {
		tmp[len] = 0;
		dprintf(fd, "%s", tmp);
	}

	return 0;
}


/*
 * Write packs installed in datadir to a file descriptor
 */
int
inspect_system(int fd, char *datadir)
{
	DIR *d;
	struct dirent *p;
	char tmp[PATH_MAX] = "";
	int meta;
	size_t len;

	if (!(d = opendir(datadir))) {
		perror(datadir);
		return -1;
	}

	while ((p = readdir(d)))
		if (strcmp(p->d_name, ".") && strcmp(p->d_name, "..")) {
			snprintf(tmp, PATH_MAX, "%s/%s/version",
				 datadir, p->d_name);

			if ((meta = open(tmp, O_RDONLY)) < 0) {
				sprintf(tmp, "(unknown)");
			} else {
				len = read(meta, tmp, 32);
				tmp[len - 1] = 0;
			}
			dprintf(fd, "%s\t%s\n", p->d_name, tmp);
		}

	closedir(d);
	return 0;
}


/*
 * Inspect the system, either by looking into a pack, checking installed
 * files, or listing packs actually installed
 */
int
inspect(char *datadir, char *packname)
{
	/* name is NULL, list packs installed */
	if (!packname)
		return inspect_system(1, datadir);

	/* otherwise, list files installed by a pack */
	return inspect_files(1, datadir, packname);
}


/*
 * Write metadata about a pack file:
 * + datadir/packname/version - version of the pack installed
 */
int
write_metadata(char *datadir, struct pack *p)
{
	int fd, r;
	struct stat st;
	char tmp[PATH_MAX];

	snprintf(tmp, PATH_MAX, "%s/%s", datadir, p->name);
	if (stat(tmp, &st) < 0) {
		if ((r = mkdir_parents(tmp, 0750)) < 0)
			return r;
	}

	snprintf(tmp, PATH_MAX, "%s/%s/version", datadir, p->name);
	if ((fd = open(tmp, O_CREAT|O_WRONLY|O_TRUNC, 0644)) < 0) {
		perror(tmp);
		return -1;
	}

	r  = write(fd, p->version, strnlen(p->version, LINE_MAX));
	r += write(fd, "\n", 1);

	if (r < 1) {
		perror(tmp);
		close(fd);
		return -1;
	}

	close(fd);
	return 0;
}


/*
 * Write an archive entry on the disk, thus creating the file
 */
int
write_entry(struct archive *a, struct archive *w)
{
	int r;
	const void *buf;
	size_t len;
	off_t off;

	for (;;) {
		r = archive_read_data_block(a, &buf, &len, &off);
		switch (r) {
		case ARCHIVE_EOF: return 0;
		case ARCHIVE_OK: break;
		default: return r;
		}

		r = archive_write_data_block(w, buf, len, off);
		if (r != ARCHIVE_OK)
			return r;
	}
}


/*
 * Extract a tarball to the given directory
 */
int
unpack(char *rootfs, char *datadir, struct pack *p)
{
	int r, fd;
	struct archive *a;
	struct archive *w;
	struct archive_entry *e;
	char cwd[PATH_MAX] = "";
	char tmp[PATH_MAX] = "";

	int mask = ARCHIVE_EXTRACT_PERM
	          |ARCHIVE_EXTRACT_SECURE_NODOTDOT;

	snprintf(tmp, PATH_MAX, "%s/%s/files", datadir, p->name);
	if ((fd = open(tmp, O_WRONLY|O_CREAT|O_TRUNC, 0644)) < 0) {
		perror(tmp);
		return -1;
	}

	a = archive_read_new();
	archive_read_support_filter_bzip2(a);
	archive_read_support_format_tar(a);

	/* try and open the tarball of our pack */
	if ((r = archive_read_open_filename(a, p->path, 0)) != ARCHIVE_OK) {
		fprintf(stderr, "%s: %s\n", p->name, archive_error_string(a));
		archive_read_free(a);
		return r;
	}

	w = archive_write_disk_new();
	archive_write_disk_set_options(w, mask);

	getcwd(cwd, PATH_MAX);
	chdir(rootfs);
	while ((r = archive_read_next_header(a, &e)) != ARCHIVE_EOF) {
		if (r != ARCHIVE_OK) {
			fprintf(stderr, "%s: %s\n", archive_entry_pathname(e),
			                            archive_error_string(a));
			break;
		}

		if ((r = archive_write_header(w, e)) != ARCHIVE_OK) {
			fprintf(stderr, "%s: %s\n", archive_entry_pathname(e),
			                            archive_error_string(w));
			break;
		}

		if ((r = write_entry(a, w)) != ARCHIVE_OK) {
			fprintf(stderr, "%s: %s\n", archive_entry_pathname(e),
			                            archive_error_string(w));
			break;
		}

		dprintf(fd, "%s\n", archive_entry_pathname(e));
		archive_write_finish_entry(w);
	}
	chdir(cwd);

	archive_write_free(w);
	archive_read_free(a);

	return r == ARCHIVE_EOF ? ARCHIVE_OK : r;
}


/*
 * Install a pack to rootfs, writing metadata to datadir
 * If overwrite is set to 1, it will overwrite all files
 */
int
install(char *rootfs, char *datadir, struct pack *p)
{
	int r;
	char tmp[PATH_MAX] = "";
	struct stat st;

	if (verbose == 1)
		printf("%s: installing to %s\n", p->name, rootfs);

	if (overwrite == 0) {
		snprintf(tmp, PATH_MAX, "%s/%s", datadir, p->name);
		if (stat(tmp, &st) == 0 && S_ISDIR(st.st_mode)) {
			fprintf(stderr, "%s: already installed\n", p->name);
			return -1;
		}

		if (verbose == 1)
			printf("%s: checking collisions\n", p->name);
		if (inspect_collision(rootfs, p) != 0)
			return -1;
	}

	if (verbose == 1)
		printf("%s: writing metadata\n", p->name);
	if (write_metadata(datadir, p) < 0) {
		perror(datadir);
		return -1;
	}

	if (verbose == 1)
		printf("%s: extracting %s\n", p->name, p->path);
	r = unpack(rootfs, datadir, p);

	if (verbose == 1)
		printf("%s: installation %s\n", p->name, r != 0 ? "failed" : "complete");

	return r;
}


/*
 * Deletes a node, be it a file or a directory.
 * In case the node doesn't exists, we're done
 */
int
delete_node(char *path)
{
	int r = 0;
	struct stat st;
	size_t len = 0;

	len = strnlen(path, PATH_MAX);
	/* remove potential trailing '/' */
	if (path[len - 1] == '/')
		path[--len] = 0;

	/*
	 * if path doesn't exist anymore, it's all good :)
	 * we use lstat here so that dangling links can be delt with too
	 */
	if (lstat(path, &st) < 0 && errno == ENOENT)
		return 0;

	if (verbose == 1)
		printf("- %s\n", path);

	if (S_ISDIR(st.st_mode) && is_empty(path) == 0) {
		if ((r = rmdir(path)) < 0) {
			perror(path);
			return r;
		}
	}

	if (!S_ISDIR(st.st_mode) && (r = unlink(path)) < 0) {
		perror(path);
		return r;
	}

	return 0;
}

/*
 * Delete all entries listed in the given file.
 * if the entry doesn't exist, it will be ignored
 */
int
delete_content(char *map, size_t size)
{
	char *path = NULL;
	size_t off;

	if (size < 1)
		return -1;

	do {
		/*
		 * it is assumed here that the file is POSIX and thus that
		 * the last char will be \n.
		 * This might sound stupid; but this is what unpack() will do.
		 * We read the mmap file backward until we encounter \n or the
		 * beginning of the file. If it's an \n, we replace it by 0 and
		 * process to the deletion of the inode, either with rmdir or
		 * unlink.
		 */
		for (off=size-1; off>0 && map[off] != '\n'; off--);
		map[off] = (off > 0 ? 0 : *map);
		path     = (off < size-1 ? &map[off] + 1 : NULL);

		if (path != NULL && strnlen(path, PATH_MAX) > 0) {
			if (delete_node(path) < 0)
				return ERR_DELETE;
		}
	} while (off > 0);

	return 0;
}


/*
 * Delete a pack from the system.
 * This will read the file datadir/name/files, change directory to rootfs
 * and call delete_content()
 */
int
delete(const char *rootfs, const char *datadir, const char *packname)
{
	int fd;
	char *addr = NULL;
	char tmp[PATH_MAX];
	struct stat st;

	snprintf(tmp, PATH_MAX, "%s/%s/files", datadir, packname);
	if (stat(tmp, &st) < 0 && errno == ENOENT) {
		fprintf(stderr, "%s: not installed\n", packname);
		return -1;
	}
	if ((fd = open(tmp, O_RDONLY)) < 0) {
		perror(tmp);
		return ERR_DELETE;
	}
	addr = mmap(0, st.st_size, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0);
	if (addr == MAP_FAILED) {
		perror(tmp);
		close(fd);
	}

	if (verbose == 1)
		printf("%s: deleting from %s\n", packname, rootfs);
	if (chdir(rootfs) < 0) {
		perror(rootfs);
		close(fd);
		return ERR_DELETE;
	}

	/* ignore errors so everything gets deleted */
	if (delete_content(addr, st.st_size) < 0) {
		fprintf(stderr, "%s: cannot remove pack\n", packname);
		close(fd);
		return ERR_DELETE;
	}
	close(fd);

	if (verbose == 1)
		printf("%s: removing metadata\n", packname);
	if (unlink(tmp) < 0) {
		perror(tmp);
		fprintf(stderr, "%s: cannot remove file list\n", packname);
		return ERR_DELETE;
	}

	snprintf(tmp, PATH_MAX, "%s/%s/version", datadir, packname);
	if (unlink(tmp) < 0) {
		perror(tmp);
		fprintf(stderr, "%s: cannot clean version\n", packname);
		return ERR_DELETE;
	}

	snprintf(tmp, PATH_MAX, "%s/%s", datadir, packname);
	return rmdir(tmp);
}


/*
 * Load a pack from a tarball and return a pack structure
 */
struct pack *
pack_load(char *path)
{
	int fd;
	struct pack *pack = NULL;
	char tmp[PATH_MAX] = "";
	char *p;

	if ((fd = open(path, O_RDONLY)) < 0) {
		perror(path);
		return NULL;
	}
	close(fd);

	if (!(pack = malloc(sizeof(struct pack)))) {
		perror(path);
		return NULL;
	}

	pack->path = NULL;
	pack->name = NULL;
	pack->version = NULL;

	pack->path = strdup(path);
	snprintf(tmp, PATH_MAX, "%s", base_name(path));

	for (p = tmp; *p && *p != PACK_SEPARATOR; p++);
	if (!*p) {
		pack_free(pack);
		return NULL;
	}
	*p = 0;
	pack->name = strdup(tmp);
	*p = PACK_SEPARATOR;

	snprintf(tmp, PATH_MAX, "%s", p + 1);
	for (p = tmp; *p && strcmp(p, PACK_EXTENSION); p++);
	if (!*p) {
		pack_free(pack);
		return NULL;
	}
	*p = 0;
	pack->version = strdup(tmp);
	*p = '.';

	return pack;
}


/*
 * Free a pack structure
 */
void
pack_free(struct pack *p)
{
	if (!p)
		return;

	if (p->path)
		free(p->path);
	if (p->name)
		free(p->name);
	if (p->version)
		free(p->version);

	free(p);
}

int
main (int argc, char **argv)
{
	int r = 0;
	char *n = NULL;
	struct pack *p = NULL;
	uint8_t action = ACTION_DEFAULT;
	char rootfs[PATH_MAX] = "";
	char datadir[PATH_MAX] = "";

	strncpy(rootfs, PACK_ROOT, PATH_MAX);
	strncat(rootfs, "/", PATH_MAX);
	strncpy(datadir, rootfs, PATH_MAX);
	strncat(datadir, PACK_DATA, PATH_MAX);

	ARGBEGIN{
	case 'a':
		action = ACTION_INSTALL;
		break;
	case 'd':
		action = ACTION_DELETE;
		break;
	case 'f':
		overwrite = 1;
		break;
	case 'i':
		action = ACTION_INSPECT;
		if (argc > 1)
			n = ARGF();
		break;
	case 'u':
		action = ACTION_UPDATE;
		break;
	case 'v':
		verbose = 1;
		break;
	default:
		action = ACTION_INVALID;
	}ARGEND;

	switch (action) {
	case ACTION_INSTALL:
		while (*argv) {
			if ((p = pack_load(*(argv++)))) {
				r += install(rootfs, datadir, p);
				pack_free(p);
			}
		}
		break;

	case ACTION_UPDATE:
		while (*argv) {
			if ((p = pack_load(*(argv++)))) {
				if (delete(rootfs, datadir, p->name) == 0)
					r += install(rootfs, datadir, p);
				pack_free(p);
			}
		}
		break;

	case ACTION_DELETE:
		while (*argv)
			r += delete(rootfs, datadir, *argv++);
		break;

	case ACTION_INSPECT:
		if (inspect(datadir, n) != 0)
			return ERR_INSPECT;
		break;

	default:
		usage(argv0);
		return ERR_INVALID_ACTION;
	}

	return r;
}
